using System;
using System.Collections.Generic;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Design;
using System.Text;
using System.Threading;
using System.Xml;

namespace Client
{
    [ServiceContract]
    public interface IBridge
    {
        [OperationContract]
        int GetDropRate();
        [OperationContract]
        void ReportReceivedCount();
        [OperationContract]
        void ReportDroppedCount();
        [OperationContract]
        void ReportForwardedCount();
    }

    public class MessageInterceptor : IMessageInterceptor
    {
        private Random randomizer = null;
        private IBridge proxy = null;
        private int dropRate = 0;

        public MessageInterceptor()
        {
            this.randomizer = new Random(DateTime.Now.Millisecond);
            this.proxy = new ChannelFactory<IBridge>("Bridge").CreateChannel();
            
        }
        
        public void ProcessSend(ref Message message)
        {
            string action = message.Headers.Action;
            Console.WriteLine(string.Format("Voting on message {0} ...",action));
            if(!(action.Contains("IImageTransfer")))
            {
                return;
            }
            this.proxy.ReportReceivedCount();
            int randomNumber = this.randomizer.Next(100);
            this.dropRate = proxy.GetDropRate();
            if((this.dropRate != 0) && (randomNumber <= this.dropRate))
            {
                Console.WriteLine("Dropping message.");
                message = null;
                this.proxy.ReportDroppedCount();
                return;
            }
            Console.WriteLine("Forwarding message.");
            this.proxy.ReportForwardedCount();
                
        }

        public void ProcessReceive(ref Message message)
        {
            
        }
    }


    public interface IMessageInterceptor
    {
        void ProcessReceive(ref Message message);
        void ProcessSend(ref Message message);
    }

    public class InterceptorBindingElement: BindingElement
    {
        IMessageInterceptor interceptor = null;

        public InterceptorBindingElement()
        {
            this.interceptor = new MessageInterceptor();
        }

        protected InterceptorBindingElement(InterceptorBindingElement other): base(other)
        {
            this.interceptor = other.interceptor;
        }

        public override IChannelFactory BuildChannelFactory(ChannelBuildContext context)
        {
            InterceptorChannelFactory result = new InterceptorChannelFactory(this.interceptor, context.Binding);
            result.InnerChannelFactory = context.BuildInnerChannelFactory();
            return result;
        }

        public override BindingElement Clone()
        {
            return new InterceptorBindingElement(this);
        }

        public override System.ServiceModel.Security.Protocols.ChannelProtectionRequirements GetProtectionRequirements()
        {
            return null;
        }
    }

    public class InterceptorSection : BindingElementExtensionSection
    {
        const string InterceptorSectionName = "interceptor";
        
        public override Type BindingElementType
        {
            get 
            { 
                return typeof(InterceptorBindingElement); 
            }
        }

        public override string ConfiguredSectionName
        {
            get 
            { 
                return InterceptorSectionName; 
            }
        }

        protected override BindingElement CreateBindingElement()
        {
            InterceptorBindingElement bindingElement = new InterceptorBindingElement();
            return bindingElement;
        }
    }

    class NullMessageInterceptor : IMessageInterceptor
    {
        public void ProcessSend(ref Message message)
        {
        }

        public void ProcessReceive(ref Message message)
        {
        }
    }

    public class InterceptorChannelFactory : ChannelFactoryBase
    {
        IMessageInterceptor interceptor;

        public InterceptorChannelFactory(IMessageInterceptor interceptor, IDefaultCommunicationTimeouts timeouts): base(timeouts)
        {
            if (interceptor == null)
            {
                this.interceptor = new NullMessageInterceptor();
            }
            else
            {
                this.interceptor = interceptor;
            }
        }

        public IMessageInterceptor Interceptor
        {
            get 
            { 
                return interceptor; 
            }
        }

        public override MessageVersion MessageVersion
        {
            get 
            { 
                return GetInnerFactorySnapshot().MessageVersion;
            }
        }

        public override string Scheme
        {
            get
            { 
                return GetInnerFactorySnapshot().Scheme;
            }
        }

        public override bool CanCreateChannel<TChannel>()
        {
            return GetInnerFactorySnapshot().CanCreateChannel<TChannel>();
        }

        protected override TChannel OnCreateChannel<TChannel>(EndpointAddress remoteAddress, Uri via)
        {
            TChannel innerChannel = this.InnerChannelFactory.CreateChannel<TChannel>(remoteAddress, via);
            if (typeof(TChannel) == typeof(IOutputChannel))
            {
                return (TChannel)(object)new InterceptorOutputChannel(this, (IOutputChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IRequestChannel))
            {
                return (TChannel)(object)new InterceptorRequestChannel(this, (IRequestChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IOutputSessionChannel))
            {
                return (TChannel)(object)new InterceptorOutputSessionChannel(this, (IOutputSessionChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IRequestSessionChannel))
            {
                return (TChannel)(object)new InterceptorRequestSessionChannel(this, 
                    (IRequestSessionChannel)innerChannel);
            }
            else if (typeof(TChannel) == typeof(IDuplexSessionChannel))
            {
                return (TChannel)(object)new InterceptorDuplexSessionChannel(this, (IDuplexSessionChannel)innerChannel, this.interceptor);
            }
            throw new ArgumentException("Channel type is not supported.", "TChannel");
        }

        protected override void OnOpen(TimeSpan timeout)
        {
            ThrowIfInnerFactoryNotSet();
            base.OnOpen(timeout);
        }

        void ThrowIfInnerFactoryNotSet()
        {
            if (this.InnerChannelFactory == null)
            {
                throw new InvalidOperationException("Inner channel factory is not set.");
            }
        }

        IChannelFactory GetInnerFactorySnapshot()
        {
            IChannelFactory innerChannelFactory = this.InnerChannelFactory;

            if (innerChannelFactory == null)
            {
                throw new InvalidOperationException("Inner channel factory is not set.");
            }

            return innerChannelFactory;
        }

        class InterceptorOutputChannel : InterceptorChannelBase<IOutputChannel>, IOutputChannel
        {
            InterceptorChannelFactory factory;

            public InterceptorOutputChannel(InterceptorChannelFactory factory, IOutputChannel innerChannel)
                : base(factory, innerChannel, factory.interceptor)
            {
                this.factory = factory;
            }

            public EndpointAddress RemoteAddress
            {
                get { return this.InnerChannel.RemoteAddress; }
            }

            public Uri Via
            {
                get { return this.InnerChannel.Via; }
            }

            public IAsyncResult BeginSend(Message message, AsyncCallback callback, object state)
            {
                return this.BeginSend(message, factory.DefaultSendTimeout, callback, state);
            }

            public IAsyncResult BeginSend(Message message, TimeSpan timeout, AsyncCallback callback, object state)
            {
                ThrowIfDisposedOrNotOpen();
                return new SendAsyncResult<IOutputChannel>(this, message, timeout, callback, state);
            }

            public void EndSend(IAsyncResult result)
            {
                SendAsyncResult<IOutputChannel>.End(result);
            }

            public void Send(Message message)
            {
                this.Send(message, factory.DefaultSendTimeout);
            }

            public void Send(Message message, TimeSpan timeout)
            {
                ThrowIfDisposedOrNotOpen();
                Interceptor.ProcessSend(ref message);
                if (message == null)
                {
                    OnDropMessage();
                }
                else
                {
                    this.InnerChannel.Send(message, timeout);
                }
            }
        }

        class InterceptorRequestChannel : InterceptorChannelBase<IRequestChannel>, IRequestChannel
        {
            InterceptorChannelFactory factory;

            public InterceptorRequestChannel(InterceptorChannelFactory factory, IRequestChannel innerChannel)
                : base(factory, innerChannel, factory.interceptor)
            {
                this.factory = factory;
            }

            public EndpointAddress RemoteAddress
            {
                get { return this.InnerChannel.RemoteAddress; }
            }

            public Uri Via
            {
                get { return this.InnerChannel.Via; }
            }

            public IAsyncResult BeginRequest(Message message, AsyncCallback callback, object state)
            {
                return this.BeginRequest(message, factory.DefaultSendTimeout, callback, state);
            }

            public IAsyncResult BeginRequest(Message message, TimeSpan timeout, AsyncCallback callback, object state)
            {
                ThrowIfDisposedOrNotOpen();
                return new RequestAsyncResult(this, message, timeout, callback, state);
            }

            public Message EndRequest(IAsyncResult result)
            {
                return RequestAsyncResult.End(result);
            }

            public Message Request(Message message)
            {
                return this.Request(message, factory.DefaultSendTimeout);
            }

            public Message Request(Message message, TimeSpan timeout)
            {
                ThrowIfDisposedOrNotOpen();
                Interceptor.ProcessSend(ref message);
                if (message == null)
                {
                    return null;
                    OnDropMessage();
                    return Message.CreateMessage(new DroppedMessageFault("Request Message dropped by interceptor."));
                }
                else
                {
                    Message reply = this.InnerChannel.Request(message, timeout);
                    Interceptor.ProcessReceive(ref reply);

                    if (reply == null)
                    {
                        OnDropMessage();
                        return Message.CreateMessage(new DroppedMessageFault("Reply Message dropped by the interceptor."));
                    }
                    else
                    {
                        return reply;
                    }
                }
            }
        }



        class InterceptorOutputSessionChannel : InterceptorOutputChannel, IOutputSessionChannel
        {
            IOutputSessionChannel innerSessionChannel;

            public InterceptorOutputSessionChannel(InterceptorChannelFactory factory, IOutputSessionChannel innerChannel)
                : base(factory, innerChannel)
            {
                this.innerSessionChannel = innerChannel;
            }

            public IOutputSession Session
            {
                get { return innerSessionChannel.Session; }
            }

            internal override void OnDropMessage()
            {
                Fault();
                innerSessionChannel.Abort();
            }
        }

        class InterceptorRequestSessionChannel : InterceptorRequestChannel, IRequestSessionChannel
        {
            IRequestSessionChannel innerSessionChannel;

            public InterceptorRequestSessionChannel(InterceptorChannelFactory factory, 
                IRequestSessionChannel innerChannel)
                : base(factory, innerChannel)
            {
                this.innerSessionChannel = innerChannel;
            }

            public IOutputSession Session
            {
                get { return innerSessionChannel.Session; }
            }

            internal override void OnDropMessage()
            {
                Fault();
                innerSessionChannel.Abort();
            }
        }

        class InterceptorDuplexChannel : InterceptorChannelBase<IDuplexChannel>, IDuplexChannel
        {
            public InterceptorDuplexChannel(ChannelManagerBase manager, IDuplexChannel innerChannel, IMessageInterceptor interceptor)
                : base(manager, innerChannel, interceptor)
            {
            }

            public EndpointAddress LocalAddress
            {
                get { return this.InnerChannel.LocalAddress; }
            }

            public EndpointAddress RemoteAddress
            {
                get { return this.InnerChannel.RemoteAddress; }
            }

            public Uri Via
            {
                get { return this.InnerChannel.Via; }
            }

            public IAsyncResult BeginReceive(AsyncCallback callback, object state)
            {
                return this.BeginReceive(DefaultReceiveTimeout, callback, state);
            }

            public IAsyncResult BeginReceive(TimeSpan timeout, AsyncCallback callback, object state)
            {
                return this.BeginTryReceive(timeout, callback, state);
            }

            public IAsyncResult BeginSend(Message message, AsyncCallback callback, object state)
            {
                return this.BeginSend(message, DefaultSendTimeout, callback, state);
            }

            public IAsyncResult BeginSend(Message message, TimeSpan timeout, AsyncCallback callback, object state)
            {
                ThrowIfDisposedOrNotOpen();
                return new SendAsyncResult<IDuplexChannel>(this, message, timeout, callback, state);
            }

            public Message EndReceive(IAsyncResult result)
            {
                Message message;
                this.EndTryReceive(result, out message);
                return message;
            }

            public void EndSend(IAsyncResult result)
            {
                SendAsyncResult<IDuplexChannel>.End(result);
            }

            public bool TryReceive(TimeSpan timeout, out Message message)
            {
                ThrowIfDisposedOrNotOpen();
                do
                {
                    if (this.InnerChannel.TryReceive(timeout, out message))
                    {
                        if (message == null)
                        {
                            return true;
                        }
                        else
                        {
                            Interceptor.ProcessReceive(ref message);
                            if (message == null)
                            {
                                OnDropMessage();
                            }
                        }
                    }
                    else
                    {
                        return false;
                    }
                } while (message == null);

                return true;
            }

            public IAsyncResult BeginTryReceive(TimeSpan timeout, AsyncCallback callback, object state)
            {
                ThrowIfDisposedOrNotOpen();
                return new TryReceiveAsyncResult<IDuplexChannel>(this, timeout, callback, state);
            }

            public bool EndTryReceive(IAsyncResult result, out Message message)
            {
                message = TryReceiveAsyncResult<IDuplexChannel>.End(result);
                return true;
            }

            public bool WaitForMessage(TimeSpan timeout)
            {
                return this.InnerChannel.WaitForMessage(timeout);
            }

            public IAsyncResult BeginWaitForMessage(TimeSpan timeout, AsyncCallback callback, object state)
            {
                return this.InnerChannel.BeginWaitForMessage(timeout, callback, state);
            }

            public bool EndWaitForMessage(IAsyncResult result)
            {
                return this.InnerChannel.EndWaitForMessage(result);
            }

            public Message Receive()
            {
                return this.Receive(DefaultReceiveTimeout);
            }

            public Message Receive(TimeSpan timeout)
            {
                Message message;
                if (this.TryReceive(timeout, out message))
                {
                    return message;
                }
                else
                {
                    throw new TimeoutException("Receive timed out.");
                }
            }

            public void Send(Message message)
            {
                this.Send(message, DefaultSendTimeout);
            }

            public void Send(Message message, TimeSpan timeout)
            {
                ThrowIfDisposedOrNotOpen();
                Interceptor.ProcessSend(ref message);
                if (message == null)
                {
                    OnDropMessage();
                }
                else
                {
                    this.InnerChannel.Send(message, timeout);
                }
            }
        }


        class InterceptorDuplexSessionChannel : InterceptorDuplexChannel, IDuplexSessionChannel
        {
            IDuplexSessionChannel innerSessionChannel;

            public InterceptorDuplexSessionChannel(ChannelManagerBase manager, IDuplexSessionChannel innerChannel, IMessageInterceptor interceptor)
                : base(manager, innerChannel, interceptor)
            {
                this.innerSessionChannel = innerChannel;
            }

            public IDuplexSession Session
            {
                get { return innerSessionChannel.Session; }
            }

            internal override void OnDropMessage()
            {
                Fault();
                innerSessionChannel.Abort();
            }
        }
        
    }

    public class InterceptorChannelBase<TChannel> : ChannelBase where TChannel : IChannel
    {
        TChannel innerChannel;
        IMessageInterceptor interceptor;

        protected InterceptorChannelBase(ChannelManagerBase manager, TChannel innerChannel, IMessageInterceptor interceptor)
            : base(manager)
        {
            this.innerChannel = innerChannel;
            this.interceptor = interceptor;
        }

        internal TChannel InnerChannel
        {
            get { return this.innerChannel; }
        }

        internal IMessageInterceptor Interceptor
        {
            get { return this.interceptor; }
        }

        protected override void OnAbort()
        {
            this.innerChannel.Abort();
        }

        protected override IAsyncResult OnBeginClose(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannel.BeginClose(timeout, callback, state);
        }

        protected override IAsyncResult OnBeginOpen(TimeSpan timeout, AsyncCallback callback, object state)
        {
            return this.innerChannel.BeginOpen(timeout, callback, state);
        }

        protected override void OnClose(TimeSpan timeout)
        {
            this.innerChannel.Close(timeout);
        }

        protected override void OnEndClose(IAsyncResult result)
        {
            this.innerChannel.EndClose(result);
        }

        protected override void OnEndOpen(IAsyncResult result)
        {
            this.innerChannel.EndOpen(result);
        }

        protected override void OnOpen(TimeSpan timeout)
        {
            this.innerChannel.Open(timeout);
        }

        internal virtual void OnDropMessage() { }
    }

    public class SendAsyncResult<TChannel> : AsyncResult where TChannel : IOutputChannel
	{
        Message message;
        TimeSpan timeout;
        InterceptorChannelBase<TChannel> channel;

        public SendAsyncResult(InterceptorChannelBase<TChannel> channel, Message message, TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.message = message;
            this.timeout = timeout;
            this.channel = channel;

            channel.Interceptor.ProcessSend(ref this.message);
            if (this.message == null)
            {
                channel.OnDropMessage();
                Complete(true);
            }
            else
            {
                channel.InnerChannel.BeginSend(this.message, timeout, new AsyncCallback(HandleCallback), null);
            }
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            try
            {
                channel.InnerChannel.EndSend(asyncResult);
                Complete(false);
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public new WaitHandle AsyncWaitHandle
        {
            get { return base.AsyncWaitHandle; }
        }

        public new bool IsCompleted
        {
            get { return base.IsCompleted; }
        }

        public new bool CompletedSynchronously
        {
            get { return base.CompletedSynchronously; }
        }

        public static void End(IAsyncResult result)
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }
            SendAsyncResult<TChannel> outputResult = result as SendAsyncResult<TChannel>;
            if (outputResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "result");
            }
            AsyncResult.End(outputResult);
        }
    }

    public abstract class AsyncResult : IAsyncResult
    {
        AsyncCallback callback;
        object state;
        bool completedSynchronously;
        bool endCalled;
        Exception exception;
        bool isCompleted;
        ManualResetEvent manualResetEvent;

        protected AsyncResult(AsyncCallback callback, object state)
        {
            this.callback = callback;
            this.state = state;
        }

        public object AsyncState
        {
            get { return state; }
        }

        public WaitHandle AsyncWaitHandle
        {
            get
            {
                if (manualResetEvent != null)
                {
                    return manualResetEvent;
                }

                lock (ThisLock)
                {
                    if (manualResetEvent == null)
                    {
                        manualResetEvent = new ManualResetEvent(isCompleted);
                    }
                }

                return manualResetEvent;
            }
        }

        public bool CompletedSynchronously
        {
            get { return completedSynchronously; }
        }

        public bool IsCompleted
        {
            get { return isCompleted; }
        }

        object ThisLock
        {
            get { return this; }
        }

        protected void Complete(bool completedSynchronously)
        {
            if (isCompleted)
            {
                throw new InvalidOperationException("AsyncResults can only be completed once.");
            }
            ManualResetEvent manualResetEvent = null;
            this.completedSynchronously = completedSynchronously;

            if (completedSynchronously)
            {
                // If we completedSynchronously, then there's no chance that the manualResetEvent was created so
                // we don't need to worry about a race
                this.isCompleted = true;
                if (this.manualResetEvent != null)
                {
                    throw new InvalidOperationException("No ManualResetEvent should be created for a synchronous AsyncResult.");
                }
            }
            else
            {
                lock (ThisLock)
                {
                    this.isCompleted = true;
                    manualResetEvent = this.manualResetEvent;
                }
            }

            try
            {
                if (callback != null)
                {
                    callback(this);
                }
                if (manualResetEvent != null)
                {
                    manualResetEvent.Set();
                }
            }
            catch (Exception unhandledException)
            {
                // The callback raising an exception is equivalent to Main raising an exception w/out a catch.
                // Queue it onto another thread and throw it there to ensure that there's no other handler in 
                // place and the default unhandled exception behavior occurs.

                // Because the stack trace gets lost on a rethrow, we're wrapping it in a generic exception
                // so the stack trace is preserved.
                unhandledException = new Exception("AsyncCallbackException", unhandledException);
                ThreadPool.UnsafeQueueUserWorkItem(new WaitCallback(RaiseUnhandledException), unhandledException);
            }
        }

        protected void Complete(bool completedSynchronously, Exception exception)
        {
            this.exception = exception;
            Complete(completedSynchronously);
        }

        protected static void End(AsyncResult asyncResult)
        {
            if (asyncResult == null)
            {
                throw new ArgumentNullException("asyncResult");
            }

            if (asyncResult.endCalled)
            {
                throw new InvalidOperationException("Async object already ended.");
            }

            asyncResult.endCalled = true;

            if (!asyncResult.isCompleted)
            {
                using (WaitHandle waitHandle = asyncResult.AsyncWaitHandle)
                {
                    waitHandle.WaitOne();
                }
            }

            if (asyncResult.exception != null)
            {
                throw asyncResult.exception;
            }
        }

        void RaiseUnhandledException(object o)
        {
            Exception exception = (Exception)o;
            throw exception;
        }
    }

    class TryReceiveAsyncResult<TChannel> : AsyncResult
        where TChannel : IInputChannel
    {
        Message message;
        TimeSpan timeout;
        InterceptorChannelBase<TChannel> channel;

        public TryReceiveAsyncResult(InterceptorChannelBase<TChannel> channel, TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.channel = channel;
            this.timeout = timeout;
            channel.InnerChannel.BeginTryReceive(timeout, new AsyncCallback(HandleCallback), null);
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            Message message;
            try
            {
                if (channel.InnerChannel.EndTryReceive(asyncResult, out message))
                {
                    if (message == null)
                    {
                        this.message = null;

                        Complete(false);
                    }
                    else
                    {
                        channel.Interceptor.ProcessReceive(ref message);

                        if (message == null)
                        {
                            channel.OnDropMessage();
                            asyncResult = channel.InnerChannel.BeginTryReceive(timeout, new AsyncCallback(HandleCallback), null);
                        }
                        else
                        {
                            this.message = message;

                            Complete(false);
                        }
                    }
                }
                else
                {
                    Complete(false, new TimeoutException("Receive request timed out."));
                }
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public static Message End(IAsyncResult asyncResult)
        {
            if (asyncResult == null)
            {
                throw new ArgumentNullException("asyncResult");
            }
            TryReceiveAsyncResult<TChannel> inputResult = asyncResult as TryReceiveAsyncResult<TChannel>;
            if (inputResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "asyncResult");
            }
            AsyncResult.End(inputResult);

            return inputResult.message;
        }

    }

    class TryReceiveRequestAsyncResult : AsyncResult
    {
        Message message;
        IRequestContext context;
        TimeSpan timeout;
        InterceptorChannelBase<IReplyChannel> channel;

        public TryReceiveRequestAsyncResult(InterceptorChannelBase<IReplyChannel> channel, TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.timeout = timeout;
            this.channel = channel;
            channel.InnerChannel.BeginTryReceiveRequest(timeout, new AsyncCallback(HandleCallback), null);
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            IRequestContext context;
            try
            {
                if (channel.InnerChannel.EndTryReceiveRequest(asyncResult, out context))
                {
                    if (context == null)
                    {
                        this.message = null;
                        this.context = null;

                        Complete(false);
                    }
                    else
                    {
                        Message result = context.RequestMessage;

                        channel.Interceptor.ProcessReceive(ref result);
                        if (result == null)
                        {
                            channel.OnDropMessage();
                            asyncResult = channel.InnerChannel.BeginTryReceiveRequest(timeout, new AsyncCallback(HandleCallback), null);
                        }
                        else
                        {
                            this.message = result;
                            this.context = context;

                            Complete(false);
                        }
                    }
                }
                else
                {
                    Complete(false, new TimeoutException("Receive request timed out."));
                }
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public static Message End(IAsyncResult asyncResult, out IRequestContext context)
        {
            if (asyncResult == null)
            {
                throw new ArgumentNullException("asyncResult");
            }
            TryReceiveRequestAsyncResult replyResult = asyncResult as TryReceiveRequestAsyncResult;
            if (replyResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "asyncResult");
            }
            AsyncResult.End(replyResult);

            context = replyResult.context;
            return replyResult.message;
        }
    }


    public class RequestAsyncResult : AsyncResult
	{
        Message message;
        TimeSpan timeout;
        InterceptorChannelBase<IRequestChannel> channel;

        public RequestAsyncResult(InterceptorChannelBase<IRequestChannel> channel, Message message, TimeSpan timeout, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.message = message;
            this.timeout = timeout;
            this.channel = channel;

            channel.Interceptor.ProcessSend(ref this.message);
            if (this.message == null)
            {
                channel.OnDropMessage();
                //this.message = Message.CreateMessage(new DroppedMessageFault("Request Message dropped by interceptor."));
                Complete(true);
            }
            else
            {
                channel.InnerChannel.BeginRequest(this.message, timeout, new AsyncCallback(HandleCallback), null);
            }
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            try
            {
                Message reply = channel.InnerChannel.EndRequest(asyncResult);
                channel.Interceptor.ProcessReceive(ref reply);
                if (reply == null)
                {
                    channel.OnDropMessage();
                    this.message = Message.CreateMessage(new DroppedMessageFault("Reply Message dropped by interceptor."));
                }
                else
                {
                    this.message = reply;
                }
                Complete(false);
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public new WaitHandle AsyncWaitHandle
        {
            get { return base.AsyncWaitHandle; }
        }

        public new bool IsCompleted
        {
            get { return base.IsCompleted; }
        }

        public new bool CompletedSynchronously
        {
            get { return base.CompletedSynchronously; }
        }

        public static Message End(IAsyncResult result)
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }
            RequestAsyncResult requestResult = result as RequestAsyncResult;
            if (requestResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "result");
            }
            AsyncResult.End(requestResult);

            return requestResult.message;
        }
    }

    public class ReplyAsyncResult : AsyncResult
	{
        Message message;
        TimeSpan timeout;
        IRequestContext innerContext;
        InterceptorChannelBase<IReplyChannel> channel;

        public ReplyAsyncResult(InterceptorChannelBase<IReplyChannel> channel, Message message,
            TimeSpan timeout, IRequestContext innerContext, AsyncCallback callback, object state)
            : base(callback, state)
        {
            this.message = message;
            this.timeout = timeout;
            this.innerContext = innerContext;
            this.channel = channel;

            channel.Interceptor.ProcessSend(ref this.message);
            if (this.message == null)
            {
                channel.OnDropMessage();
                this.message = Message.CreateMessage(new DroppedMessageFault("Reply Message dropped by interceptor."));
                Complete(true);
            }
            else
            {
                innerContext.BeginReply(this.message, new AsyncCallback(HandleCallback), null);
            }
        }

        void HandleCallback(IAsyncResult asyncResult)
        {
            try
            {
                innerContext.EndReply(asyncResult);
                Complete(false);
            }
            catch (Exception e)
            {
                Complete(false, e);
            }
        }

        public new WaitHandle AsyncWaitHandle
        {
            get { return base.AsyncWaitHandle; }
        }

        public new bool IsCompleted
        {
            get { return base.IsCompleted; }
        }

        public new bool CompletedSynchronously
        {
            get { return base.CompletedSynchronously; }
        }

        public static void End(IAsyncResult result)
        {
            if (result == null)
            {
                throw new ArgumentNullException("result");
            }
            ReplyAsyncResult requestResult = result as ReplyAsyncResult;
            if (requestResult == null)
            {
                throw new ArgumentException("Invalid AsyncResult", "result");
            }
            AsyncResult.End(requestResult);
        }
	}

    public class DroppedMessageFault : MessageFault
    {
        FaultReason faultReason;
        FaultCode faultCode;

        public DroppedMessageFault(string reason)
        {
            faultReason = new FaultReason(reason);
            faultCode = new FaultCode("DroppedMessageFault");
        }

        public override FaultCode Code
        {
            get { return faultCode; }
        }
        
        public override bool HasDetail
        {
            get { return false; }
        }

        public override FaultReason Reason
        {
            get { return faultReason; }
        }
        /* November*/
        protected override void OnWriteDetailContents(XmlDictionaryWriter writer)
        {
            throw new NotImplementedException("The method or operation is not implemented.");
        }
       
    }
}
